/**********************************************************************
		NOTEMIDI - module for serial port MIDI devices

		(C) 2001 by Michael Minn - michael@michaelminn.com

   Kernel source MUST be loaded on the system in order to compile this
   module.  Because modules cannot be dynamically loaded into the Linux
   Sound System (OSS), this module uses functions from the kernel source
   tree that enables this module to be listed on /dev/sndstat.

   See /usr/src/linux/drivers/char/serial.c for details on the handling of
   the serial port.

   printk messages are written to /var/log/messages

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

***************************************************************************/

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/sound.h>
#include <linux/poll.h>
#include <linux/tty.h>
#include <sys/syscall.h>
#include <sys/errno.h>
#include <asm/uaccess.h>

/* These include files are part of the kernel source tree */
#include <drivers/sound/sound_config.h>
#include <drivers/sound/dev_table.h>

#define SERIAL_PORT_DEVICE_NAME "/dev/ttyS0"

#ifndef NULL
#define NULL 0
#endif

static struct file *serial_port_file = NULL;
static int notemidi_device_id = 0;
static int ms124w_device_id = 0;

/*************************************************************************/

static int notemidi_open (int device, int mode, void (*input) (int dev, unsigned char data), void (*output) (int dev))
{
	return 0;
}

static void notemidi_close (int device)
{
}

static int notemidi_out (int device, unsigned char midi_byte)
{
	int status = 0;
	struct tty_struct *tty = (struct tty_struct *)serial_port_file->private_data;
	
	if (tty == NULL)
		printk(KERN_INFO "NOTEMIDI: NULL file private data TTY pointer\n");
		
	else if (tty->driver.write == NULL)
		printk(KERN_INFO "NOTEMIDI: NULL TTY driver write function\n");
	
	else if ((status = tty->driver.write(tty, 0, &midi_byte, 1)) < 1)
		printk(KERN_INFO "NOTEMIDI: write failure: device %d, byte %x, status %d\n", device, (int) midi_byte, status);

	/***
	if ((status = serial_port_file->f_op->write(serial_port_file, &midi_byte, 1, &serial_port_file->f_pos)) < 1)
		printk(KERN_INFO "NOTEMIDI: write failure - device %d, byte %x, status %d\n", device, (int) midi_byte, status);
	***/
			
	return status;
}

/************************************************
MS-124 - one Burst Byte preceeds each data byte
lower 4 bits always 8
upper 4 bits correspond to output port
0x18 = port 1
0x28 = port 2
0x48 = port 3
0x88 = port 4
0xf8 = all four ports simultaneously
************************************************/

static int ms124w_out (int device, unsigned char midi_byte)
{
	char burst_byte = 0xf8;
	int status = 0;
	/* see /usr/include/linux/tty.h and tty_driver.h */
     	
	struct tty_struct *tty = (struct tty_struct *)serial_port_file->private_data;
	
	if (tty == NULL)
		printk(KERN_INFO "NOTEMIDI: NULL file private data TTY pointer\n");
		
	else if (tty->driver.write == NULL)
		printk(KERN_INFO "NOTEMIDI: NULL TTY driver write function\n");
	
	else if ((status = tty->driver.write(tty, 0, &burst_byte, 1)) < 1)
     	printk(KERN_INFO "NOTEMIDI: burst write failure: ms124 device %d, status %d\n", device, status);
	
	else if ((status = tty->driver.write(tty, 0, &midi_byte, 1)) < 1)
		printk(KERN_INFO "NOTEMIDI: write failure: ms124 device %d, byte %x, status %d\n", device, (int) midi_byte, status);

	/**		
	if ((status = serial_port_file->f_op->write(serial_port_file, &burst_byte, 1, &serial_port_file->f_pos)) < 1)
     	printk(KERN_INFO "NOTEMIDI: burst write failure: ms124 device %d, status %d\n", device, status);
	
	if ((status = serial_port_file->f_op->write(serial_port_file, &midi_byte, 1, &serial_port_file->f_pos)) < 1)
		printk(KERN_INFO "NOTEMIDI: write failure: ms124 device %d, byte %x, status %d\n", device, (int) midi_byte, status);
	**/
	
	return status;
}

static int notemidi_start_read (int device)
{
	return 0;
}

static int notemidi_end_read (int device)
{
	return 0;
}

static int notemidi_ioctl (int dev, unsigned cmd, caddr_t arg)
{
	return -EPERM;
}

static struct midi_operations notemidi_operations =
{
	{"NOTEMIDI", 0, 0, 0},
	NULL,
	{0},
	notemidi_open,
	notemidi_close,
	notemidi_ioctl,
	notemidi_out,
	notemidi_start_read,
	notemidi_end_read,
	NULL,
	NULL,
	NULL,
	NULL
};

static struct midi_operations ms124w_operations =
{
	{"MS-124W", 0, 0, 0},
	NULL,
	{0},
	notemidi_open,
	notemidi_close,
	notemidi_ioctl,
	ms124w_out,
	notemidi_start_read,
	notemidi_end_read,
	NULL,
	NULL,
	NULL,
	NULL
};

/********************************************************/

int open_serial_port(char *serial_port_device_name)
{
	struct tty_struct *tty;
	struct termios old_termios;
	
	if (IS_ERR(PTR_ERR(serial_port_file = filp_open(serial_port_device_name, O_WRONLY, 0))))
		printk(KERN_INFO "NOTEMIDI: Failure %d opening serial port %s\n", (int) PTR_ERR(serial_port_file), serial_port_device_name);
		
	else if ((tty = (struct tty_struct *)serial_port_file->private_data) == NULL)
		printk(KERN_INFO "NOTEMIDI: NULL file private data TTY pointer\n");
	
	else if (tty->driver.set_termios == NULL)
		printk(KERN_INFO "NOTEMIDI: NULL file set_termios function\n");

	else if (tty->termios == NULL)
		printk(KERN_INFO "NOTEMIDI: NULL file termios structure\n");

	else
	{
		old_termios = *tty->termios;
		tty->termios->c_lflag = 0;
		tty->termios->c_lflag = NOFLSH;
		tty->termios->c_iflag = IGNBRK | IGNPAR;
		tty->termios->c_oflag = 0;
		tty->termios->c_cflag =  B38400 | CREAD | CSIZE | CS8 | CLOCAL | HUPCL;
		tty->termios->c_cc[VEOL] = 0; /* '\r'; */
		tty->termios->c_cc[VERASE] = 0;
		tty->termios->c_cc[VKILL] = 0;
		tty->termios->c_cc[VMIN] = 0;
		tty->termios->c_cc[VTIME] = 0;
		(*tty->driver.set_termios)(tty, &old_termios);
	}
		
	return 0;
}


/********************************************************/

int init_module(void)
{
	printk(KERN_INFO "NOTEMIDI: Linux Serial Port MIDI interface module v2001.05.21 - loading\n");
	
	open_serial_port(SERIAL_PORT_DEVICE_NAME);
	
	if ((notemidi_device_id = sound_alloc_mididev()) < 0)
		printk(KERN_ERR "NOTEMIDI: sound_alloc_mididev() failed\n");
	
	else if ((ms124w_device_id = sound_alloc_mididev()) < 0)
		printk(KERN_ERR "NOTEMIDI: MS-124W sound_alloc_mididev() failed\n");
	
	else
	{
		printk(KERN_ERR "NOTEMIDI: main device id %d, ms124w id %d\n", notemidi_device_id, ms124w_device_id);
		midi_devs[notemidi_device_id] = &notemidi_operations;
		midi_devs[ms124w_device_id] = &ms124w_operations;
	}
	
	return 0;
}

int cleanup_module(void)
{
	if (! IS_ERR(PTR_ERR(serial_port_file)))
		filp_close(serial_port_file, 0);
	
	sound_unload_mididev(notemidi_device_id);
	sound_unload_mididev(ms124w_device_id);

	printk(KERN_INFO "NOTEMIDI: Unloading module\n");
	
	return 0;
}
